/*
 * This code is part of the programming language flup
 * flup comes with ABSOLUTELY NO WARRANTY and is licensed under AGPL-3.0 or later.
 *    Copyright (C) 2024  VegOwOtenks
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include "builtin-functions.h"

#include "value.h"

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>

int BuiltinFunction_Equality(CallFrame* top_frame)
{
	Value v1;
	Value v2;

	CallFrame_StackPop(top_frame, &v1);
	if (CallFrame_StackPop(top_frame, &v2) != EXIT_SUCCESS) {
		// TODO: Error message
		return EXIT_FAILURE;
	}

	Value result;
	result.type = VALUETYPE_BOOLEAN;
	result.get.boolean = Value_Equal(&v1, &v2);

	top_frame->instruction_pointer++;
	return CallFrame_StackPush(top_frame, &result);

}

int BuiltinFunction_Plus(CallFrame* top_frame)
{
	Value v1;
	Value v2;

	CallFrame_StackPop(top_frame, &v1);
	if (CallFrame_StackPop(top_frame, &v2) != EXIT_SUCCESS) {
		// TODO: Error message
		return EXIT_FAILURE;
	}
	if (v1.type != v2.type) {
		// TODO: Error message
		return EXIT_FAILURE;
	}

	Value result;
	result.type = v1.type;
	switch (v1.type) {
		case VALUETYPE_INT64:
			result.get.i64 = v2.get.i64 + v1.get.i64;
			break;
		case VALUETYPE_DOUBLE:
			result.get.f64 = v2.get.f64 + v1.get.f64;
			break;
		case VALUETYPE_BOOLEAN:
			result.get.boolean = v2.get.boolean + v1.get.boolean;
			break;
		case VALUETYPE_OBJECT:
			return EXIT_FAILURE;
		case VALUETYPE_BYTE:
			return v1.get.byte + v2.get.byte;
	}

	top_frame->instruction_pointer++;
	return CallFrame_StackPush(top_frame, &result);
}

int BuiltinFunction_Multiply(CallFrame* top_frame)
{
	Value v1;
	Value v2;

	CallFrame_StackPop(top_frame, &v1);
	if (CallFrame_StackPop(top_frame, &v2) != EXIT_SUCCESS) {
		// TODO: Error message
		return EXIT_FAILURE;
	}
	if (v1.type != v2.type) {
		// TODO: Error message
		return EXIT_FAILURE;
	}

	Value result;
	result.type = v1.type;
	switch (v1.type) {
		case VALUETYPE_INT64:
			result.get.i64 = v2.get.i64 * v1.get.i64;
			break;
		case VALUETYPE_DOUBLE:
			result.get.f64 = v2.get.f64 * v1.get.f64;
			break;
		case VALUETYPE_BOOLEAN:
			result.get.boolean = v2.get.boolean * v1.get.boolean;
			break;
		case VALUETYPE_OBJECT:
			return EXIT_FAILURE;
		case VALUETYPE_BYTE:
			return v1.get.byte * v2.get.byte;
	}

	top_frame->instruction_pointer++;
	return CallFrame_StackPush(top_frame, &result);
}

int BuiltinFunction_Minus(CallFrame* top_frame)
{
	Value v1;
	Value v2;

	CallFrame_StackPop(top_frame, &v1);
	if (CallFrame_StackPop(top_frame, &v2) != EXIT_SUCCESS) {
		// TODO: Error message
		return EXIT_FAILURE;
	}
	if (v1.type != v2.type) {
		// TODO: Error message
		return EXIT_FAILURE;
	}

	Value result;
	result.type = v1.type;
	switch (v1.type) {
		case VALUETYPE_INT64:
			result.get.i64 = v2.get.i64 - v1.get.i64;
			break;
		case VALUETYPE_DOUBLE:
			result.get.f64 = v2.get.f64 - v1.get.f64;
			break;
		case VALUETYPE_BOOLEAN:
			result.get.boolean = v2.get.boolean - v1.get.boolean;
			break;
		case VALUETYPE_OBJECT:
			return EXIT_FAILURE;
		case VALUETYPE_BYTE:
			return v1.get.byte - v2.get.byte;
	}

	top_frame->instruction_pointer++;
	return CallFrame_StackPush(top_frame, &result);
}

int BuiltinFunction_Print(CallFrame* top_frame)
{
	Value v1;

	if (CallFrame_StackPop(top_frame, &v1) != EXIT_SUCCESS) {
		return EXIT_FAILURE;
	}

	switch (v1.type) {
		case VALUETYPE_INT64:
			printf("%" PRIi64, v1.get.i64);
			break;
		case VALUETYPE_DOUBLE:
			printf("%f", v1.get.f64);
			break;
		case VALUETYPE_BOOLEAN:
			if (v1.get.boolean) {
				printf("true");
			} else {
				printf("false");
			}
			break;
		case VALUETYPE_OBJECT:
			if (v1.get.object != NULL) {
				StringView type_name_view = v1.get.object->type->name;
				char type_name[type_name_view.length];
				type_name[type_name_view.length] = '\0';
				StringView_Paste(type_name, type_name_view);
				printf("object (%s) at %p\n", type_name, (void*) v1.get.object);
			} else {
				puts("null");
			}
			break;
		case VALUETYPE_BYTE:
			printf("%c", v1.get.byte);
			break;
	}

	return EXIT_SUCCESS;
}

int BuiltinFunction_PrintLine(CallFrame* top_frame)
{
	if (BuiltinFunction_Print(top_frame)) {
		return EXIT_FAILURE;
	} else {
		puts("");
		return EXIT_SUCCESS;
	}
}

int BuiltinFunction_Duplicate(CallFrame* top_frame)
{
	Value v1;

	if (CallFrame_StackPop(top_frame, &v1) != EXIT_SUCCESS) {
		return EXIT_FAILURE;
	}

	if (CallFrame_StackPush(top_frame, &v1) || CallFrame_StackPush(top_frame, &v1)) {
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
